@@ -1,8 +1,8 @@
-__version__ = "0.4.1"
+__version__ = "0.4.2"
 
 from pathlib import Path
 
-from ugrd import ValidationError
+from ugrd.exceptions import ValidationError
 from zenlib.util import contains
 
 
@@ -31,8 +31,11 @@ def _check_in_file(self, file, lines):
     with open(file, "r") as f:
         file_lines = f.readlines()
 
+    stripped_lines = [line.strip() for line in file_lines]
+
     for check_line in lines:
-        if check_line not in file_lines:
+        if check_line not in file_lines and check_line not in stripped_lines:
+            self.logger.info(f"File lines: {file_lines}")
             raise ValueError("Failed to find line '%s' in file '%s'" % (check_line, file))
 
 
@@ -70,3 +73,33 @@ def check_included_or_mounted(self):
                 self.logger.error("File detected under mount '%s' but is not present: %s" % (mountpoint, file))
 
     return "All included files were found."
+
+
+def check_init_order(self):
+    """Ensures that init functions are ordered respecting import_order"""
+    for hook, hook_funcs in self["imports"].items():  # Iterate through all imported functions
+        if hook == "custom_init":
+            continue  # Only one function should be in herE
+        hook_funcs = [func.__name__ for func in hook_funcs]
+        a = self["import_order"].get("after", {})
+        after = {k: v for k, v in a.items() if k in hook_funcs and any(subv in hook_funcs for subv in v)}
+        b = self["import_order"].get("before", {})
+        before = {k: v for k, v in b.items() if k in hook_funcs and any(subv in hook_funcs for subv in v)}
+        for func, targets in before.items():
+            for target in targets:
+                func_index = hook_funcs.index(func)
+                try:
+                    target_index = hook_funcs.index(target)
+                except ValueError:
+                    continue  # Ignore targets that are not imported
+                if func_index > target_index:
+                    raise ValidationError("[%s] Function must be before: %s" % (func, target))
+        for func, targets in after.items():
+            for target in targets:
+                func_index = hook_funcs.index(func)
+                try:
+                    target_index = hook_funcs.index(target)
+                except ValueError:
+                    continue  # Ignore targets that are not imported
+                if func_index < target_index:
+                    raise ValidationError("[%s] Function must be after: %s" % (func, target))
