@@ -1,21 +1,43 @@
 __author__ = "desultory"
-__version__ = "4.1.3"
+__version__ = "4.7.0"
 
+from os import environ, makedev, mknod, uname
 from pathlib import Path
 from shutil import rmtree, which
+from stat import S_IFCHR
+from subprocess import run
 from typing import Union
 
-from ugrd import AutodetectError, ValidationError
+from ugrd.exceptions import AutodetectError, ValidationError
 from zenlib.types import NoDupFlatList
-from zenlib.util import colorize, contains, unset
+from zenlib.util import colorize as c_
+from zenlib.util import contains, unset
 
 
-def detect_tmpdir(self) -> None:
+def _has_zstd() -> bool:
+    from importlib.util import find_spec
+
+    if find_spec("zstandard"):
+        return True
+    return False
+
+
+class DecompressorError(Exception):
+    """Custom exception for decompressor errors."""
+
+    pass
+
+
+class LDConfigError(Exception):
+    """Custom exception for ldconfig errors."""
+
+    pass
+
+
+def get_tmpdir(self) -> None:
     """Reads TMPDIR from the environment, sets it as the temporary directory."""
-    from os import environ
-
     if tmpdir := environ.get("TMPDIR"):
-        self.logger.info("Detected TMPDIR: %s" % (colorize(tmpdir, "cyan")))
+        self.logger.info("Detected TMPDIR: %s" % (c_(tmpdir, "cyan")))
         self["tmpdir"] = Path(tmpdir)
 
 
@@ -24,28 +46,35 @@ def _get_shell_path(self, shell_name) -> Path:
     if shell := which(shell_name):
         return Path(shell).resolve()
     else:
-        raise AutodetectError(f"Shell '{shell_name}' not found.")
+        raise AutodetectError(f"Shell not found: {shell_name}")
 
 
 def get_shell(self) -> None:
     """Gets the shell, uses /bin/sh if a shell is not set"""
     if shell := self["shell"]:
-        self.logger.info("Using shell: %s", colorize(shell, "blue", bright=True))
+        self.logger.info("Using shell: %s", c_(shell, "blue", bright=True))
         self["binaries"] = shell
         shell_path = _get_shell_path(self, shell)
         self["symlinks"]["shell"] = {"target": "/bin/sh", "source": shell_path}
     else:
-        self.logger.info("Using default shell: %s", colorize("/bin/sh", "cyan"))
+        self.logger.info("Using default shell: %s", c_("/bin/sh", "cyan"))
         self["binaries"] = "/bin/sh"  # This should pull /bin/sh and the target if it's a symlink
 
 
 @contains("clean", "Skipping cleaning build directory", log_level=30)
 def clean_build_dir(self) -> None:
-    """Cleans the build directory."""
+    """Cleans the build directory.
+    Ensures there are no active mounts in the build directory.
+    """
     build_dir = self._get_build_path("/")
 
+    if any(mount.startswith(str(build_dir)) for mount in self["_mounts"]):
+        self.logger.critical("Mount detected in build directory, stopping: %s" % build_dir)
+        self.logger.critical("Active mounts: %s" % self["_mounts"])
+        exit(1)
+
     if build_dir.is_dir():
-        self.logger.warning("Cleaning build directory: %s" % colorize(build_dir, "yellow"))
+        self.logger.warning("Cleaning build directory: %s" % c_(build_dir, "yellow"))
         rmtree(build_dir)
     else:
         self.logger.info("Build directory does not exist, skipping cleaning: %s" % build_dir)
@@ -57,7 +86,7 @@ def generate_structure(self) -> None:
         self._mkdir(subdir)
 
 
-def add_conditional_dependencies(self) -> None:
+def get_conditional_dependencies(self) -> None:
     """Adds conditional dependencies to the dependencies list.
     Keys are the dependency, values are a tuple of the condition type and value.
     """
@@ -77,26 +106,46 @@ def add_conditional_dependencies(self) -> None:
                 unset(condition_value)(add_dep(dependency))
 
 
-def calculate_dependencies(self, binary: str) -> list[Path]:
-    """Calculates the dependencies of a binary using lddtree
-    :param binary: The binary to calculate dependencies for
-    :return: A list of dependency paths
-    """
-    from subprocess import run
+def _determine_interpreter(self, binary: Path) -> str:
+    """Checks the shebang of a file, returning the interpreter if it exists."""
+    with binary.open("rb") as f:
+        try:
+            first_line = f.readline().decode("utf-8").strip()
+        except UnicodeDecodeError:
+            return self.logger.debug(f"Binary is not a text file, skipping shebang check: {c_(binary, 'yellow')}")
 
+        if first_line.startswith("#!"):
+            interpreter = first_line[2:].split()[0]
+            self.logger.debug(f"[{binary}] Interpreter found: {c_(interpreter, 'green')}")
+            return interpreter
+        else:
+            self.logger.log(5, "No shebang found in: %s" % binary)
+
+
+def calculate_dependencies(self, binary: str) -> list[Path]:
+    """Calculates the dependencies of a binary using lddtree.
+
+    Additionally, pulls the interpreter if defined in the binary's shebang.
+
+    Returns a list of Path objects for each dependency.
+    """
     binary_path = which(binary)
     if not binary_path:
-        raise AutodetectError("'%s' not found in PATH" % binary)
+        raise AutodetectError(f"Binary not found not found in PATH: {binary}")
 
     binary_path = Path(binary_path)
+    if interpreter := _determine_interpreter(self, binary_path):
+        if interpreter not in self["binaries"]:
+            self.logger.info(f"[{c_(binary, 'blue')}] Adding interpreter to binaries: {c_(interpreter, 'cyan')}")
+            self["binaries"] = interpreter
+        else:
+            self.logger.debug(f"Interpreter already in binaries list, skipping: {c_(interpreter, 'yellow')}")
 
     self.logger.debug("Calculating dependencies for: %s" % binary_path)
     dependencies = run(["lddtree", "-l", str(binary_path)], capture_output=True)
 
     if dependencies.returncode != 0:
-        self.logger.warning(
-            "Unable to calculate dependencies for: %s" % colorize(binary, "red", bold=True, bright=True)
-        )
+        self.logger.warning("Unable to calculate dependencies for: %s" % c_(binary, "red", bold=True, bright=True))
         raise AutodetectError("Unable to resolve dependencies, error: %s" % dependencies.stderr.decode("utf-8"))
 
     dependency_paths = []
@@ -110,6 +159,23 @@ def calculate_dependencies(self, binary: str) -> list[Path]:
     return dependency_paths
 
 
+def find_library(self, library: str) -> None:
+    """Given a library file name, searches for it in the library paths, adds it to the dependencies list."""
+    search_paths = set(self["library_paths"]) | {"/lib", "/lib64", "/usr/lib", "/usr/lib64"}
+
+    for path in search_paths:
+        lib_path = Path(path).joinpath(library)
+        if lib_path.exists():
+            self.logger.info("[%s] Found library file: %s" % (library, c_(lib_path, "cyan")))
+            return lib_path
+        # Attempt to find the library with a .so extension
+        lib_path = lib_path.with_suffix(".so")
+        if lib_path.exists():
+            self.logger.info("[%s] Found library file: %s" % (library, c_(lib_path, "cyan")))
+            return lib_path
+    raise AutodetectError("Library not found: %s" % library)
+
+
 @contains("merge_usr", "Skipping /usr merge", log_level=30)
 def handle_usr_symlinks(self) -> None:
     """
@@ -128,8 +194,9 @@ def handle_usr_symlinks(self) -> None:
     for target, source in symlinks:
         host_path = Path("/").joinpath(target)
         if host_path.is_dir() and not host_path.is_symlink():
-            self.logger.warning("Host path is a directory: %s" % host_path)
+            self.logger.warning("Host path is a directory, skipping symlink creation: %s" % host_path)
             self.logger.warning("Set `merge_usr = false` to disable /usr merge.")
+            continue
         self._symlink(source, target)
 
 
@@ -146,34 +213,65 @@ def deploy_dependencies(self) -> None:
         self._copy(dependency)
 
 
+def _deploy_compressed(self, compression_type: str, decompressor, compression_extensions=None) -> None:
+    """Decompresses all dependencies of the specified compression type into the build directory.
+    Remove the compression extension if there is a match between compression_extensions and the file name.
+    """
+    compression_extensions = compression_extensions or [f".{compression_type}"]
+
+    for dependency in self[f"{compression_type}_dependencies"]:
+        self.logger.debug(f"[{compression_type}] Decompressing: {dependency}")
+
+        for extension in compression_extensions:
+            if str(dependency).endswith(extension):
+                break
+        else:
+            self.logger.warning(f"[{compression_type}] Dependency missing extension: {dependency}")
+            extension = ""
+
+        self.logger.debug(f"[{compression_type}] Found compressed file: {dependency}")
+        # Replace the extension, do nothing if there is no match
+        out_path = self._get_build_path(str(dependency).replace(extension, ""))
+        if not out_path.parent.is_dir():
+            self.logger.debug(f"Creating parent directory: {out_path.parent}")
+            self._mkdir(out_path.parent, resolve_build=False)
+
+        try:
+            data = decompressor(dependency.read_bytes())
+        except Exception as e:
+            raise DecompressorError(f"[{compression_type}] Unable to decompress ependency: {dependency} ({e})")
+
+        with out_path.open("wb") as out_file:
+            out_file.write(data)
+            self.logger.info(
+                f"[{c_(compression_type, 'green', bright=True)}] Decompressed {c_(dependency, 'blue')} -> {c_(out_path, 'green')}"
+            )
+
+
+@contains("xz_dependencies", "No xz dependencies defined, skipping.", log_level=10)
 def deploy_xz_dependencies(self) -> None:
     """Decompresses all xz dependencies into the build directory."""
     from lzma import decompress
 
-    for xz_dependency in self["xz_dependencies"]:
-        self.logger.debug("[xz] Decompressing: %s" % xz_dependency)
-        out_path = self._get_build_path(str(xz_dependency).replace(".xz", ""))
-        if not out_path.parent.is_dir():
-            self.logger.debug("Creating parent directory: %s" % out_path.parent)
-            self._mkdir(out_path.parent, resolve_build=False)
-        with out_path.open("wb") as out_file:
-            out_file.write(decompress(xz_dependency.read_bytes()))
-            self.logger.info("[xz] Decompressed '%s' to: %s" % (xz_dependency, out_path))
+    _deploy_compressed(self, "xz", decompress)
 
 
+@contains("zstd_dependencies", "No zstd dependencies defined, skipping.", log_level=10)
+def deploy_zstd_dependencies(self) -> None:
+    """Decompresses all zstd dependencies into the build directory.
+    Entries should only be added to zstd_dependencies if the zstandard library is available.
+    """
+    from zstandard import decompress
+
+    _deploy_compressed(self, "zstd", decompress, compression_extensions=[".zst", ".zstd"])
+
+
+@contains("gz_dependencies", "No gz dependencies defined, skipping.", log_level=10)
 def deploy_gz_dependencies(self) -> None:
     """Decompresses all gzip dependencies into the build directory."""
     from gzip import decompress
 
-    for gz_dependency in self["gz_dependencies"]:
-        self.logger.debug("[gz] Decompressing: %s" % gz_dependency)
-        out_path = self._get_build_path(str(gz_dependency).replace(".gz", ""))
-        if not out_path.parent.is_dir():
-            self.logger.debug("Creating parent directory: %s" % out_path.parent)
-            self._mkdir(out_path.parent, resolve_build=False)
-        with out_path.open("wb") as out_file:
-            out_file.write(decompress(gz_dependency.read_bytes()))
-            self.logger.info("[gz] Decompressed '%s' to: %s" % (gz_dependency, out_path))
+    _deploy_compressed(self, "gz", decompress)
 
 
 def deploy_copies(self) -> None:
@@ -190,12 +288,10 @@ def deploy_symlinks(self) -> None:
         self._symlink(symlink_parameters["source"], symlink_parameters["target"])
 
 
+@contains("nodes", "Skipping device node creation, no nodes are defined.")
 @contains("make_nodes", "Skipping real device node creation with mknod, as make_nodes is not specified.", log_level=20)
 def deploy_nodes(self) -> None:
     """Generates specified device nodes."""
-    from os import makedev, mknod
-    from stat import S_IFCHR
-
     for node, config in self["nodes"].items():
         node_path_abs = Path(config["path"])
 
@@ -213,39 +309,79 @@ def deploy_nodes(self) -> None:
             raise e
 
 
-@contains("find_libgcc", "Skipping libgcc_s dependency resolution", log_level=20)
-def find_libgcc(self) -> None:
-    """Finds libgcc.so, adds a 'dependencies' item for it.
-    Adds the parent directory to 'library_paths'
-    """
-    from pathlib import Path
-
-    musl_warning = False
+def _get_ldconfig(self) -> list:
+    """Returns the ldconfig command if it exists, otherwise raise an LDConfigError."""
     try:
         cmd = self._run(["ldconfig", "-p"], fail_silent=True, fail_hard=False)
     except FileNotFoundError:
-        musl_warning = True
+        self.logger.warning("If musl libc is being used, `musl_libc = true` should be set in the config.")
+        raise LDConfigError("ldconfig not found, unable to resolve libraries.")
 
-    if not musl_warning and b"Unimplemented option: -p" in cmd.stderr:  # Probably musl libc
-        musl_warning = True
-
-    if musl_warning:
-        self.logger.warning("This check can be disabled by setting `find_libgcc = false` in the configuration.")
-        return self.logger.warning("Unable to run ldconfig -p, if GCC is being used, this is fatal!")
+    if b"Unimplemented option: -p" in cmd.stderr:  # Probably musl libc
+        self.logger.warning("If musl libc is being used, `musl_libc = true` should be set in the config.")
+        raise LDConfigError("ldconfig -p not supported, musl libc is likely in use.")
 
     if cmd.returncode != 0:
-        return self.logger.critical("Unable to run ldconfig -p, if GCC is being used, this is fatal!")
+        raise LDConfigError("ldconfig failed to run: %s" % cmd.stderr.decode("utf-8"))
 
-    ldconfig = cmd.stdout.decode("utf-8").splitlines()
+    return cmd.stdout.decode("utf-8").splitlines()
+
+
+@unset("musl_libc", "Skipping libgcc_s dependency resolution, musl_libc is enabled.", log_level=20)
+@contains("find_libgcc", "Skipping libgcc_s dependency resolution", log_level=20)
+def autodetect_libgcc(self) -> None:
+    """Finds libgcc.so, adds a 'dependencies' item for it.
+    Adds the parent directory to 'library_paths'
+    """
+    try:
+        ldconfig = _get_ldconfig(self)
+    except LDConfigError:
+        self.logger.warning("Unable to run ldconfig -p, if glibc is being used, this is fatal!")
+        self.logger.warning("If libgcc_s is not required, set `find_libgcc = false` in the configuration.")
+        return
 
     libgcc = [lib for lib in ldconfig if "libgcc_s" in lib and "(libc6," in lib][0]
     source_path = Path(libgcc.partition("=> ")[-1])
-    self.logger.info("Source path for libgcc_s: %s" % source_path)
+    self.logger.info(f"Source path for libgcc_s: {c_(source_path, 'green')}")
 
     self["dependencies"] = source_path
     self["library_paths"] = str(source_path.parent)
 
 
+def autodetect_musl(self) -> None:
+    """Looks for /etc/ld-musl-<arch>.path
+    If it exists, adds it to the dependencies list.
+    """
+    arch = uname().machine
+    musl_path = Path(f"/etc/ld-musl-{arch}.path")
+    if musl_path.exists():
+        self.logger.info(f"Detected musl search path: {c_(musl_path, 'green')}")
+        self["dependencies"] = musl_path
+    elif self["musl_libc"]:
+        raise AutodetectError("Musl libc is enabled, but the musl search path was not found: %s" % musl_path)
+
+
+@unset("musl_libc", "Skipping ld.so.cache regeneration, musl_libc is enabled.", log_level=30)
+def regen_ld_so_cache(self) -> None:
+    """
+    Checks if a 'real' ldconfig is available, if so, regenerates the ld.so.cache file in the build dir.
+    Uses defined library paths to generate the config file.
+
+    If ldconfig is not available, _get_ldconfiig will warn about setting `musl_libc` to true.
+    Here, warn about this and that this is a fatal error if glibc is being used.
+    """
+    try:
+        _get_ldconfig(self)
+    except LDConfigError:
+        return self.logger.warning("Unable to run ldconfig -p, if glibc is being used, this is fatal!")
+
+    self.logger.info("Regenerating ld.so.cache")
+    self._write("etc/ld.so.conf", self["library_paths"])
+    build_path = self._get_build_path("/")
+    self._run(["ldconfig", "-r", str(build_path)])
+    self["check_included_or_mounted"] = "etc/ld.so.cache"
+
+
 def _process_out_file(self, out_file: str) -> None:
     """Processes the out_file.
 
@@ -256,7 +392,7 @@ def _process_out_file(self, out_file: str) -> None:
     out_file = str(out_file)
     if out_file == "./" or out_file == ".":
         current_dir = Path(".").resolve()
-        self.logger.info("Setting out_dir to current directory: %s" % colorize(current_dir, "cyan", bold=True))
+        self.logger.info("Setting out_dir to current directory: %s" % c_(current_dir, "cyan", bold=True))
         self["out_dir"] = current_dir
         return
 
@@ -276,7 +412,7 @@ def _process_out_file(self, out_file: str) -> None:
 
     if str(out_file.parent) != ".":  # If the parent isn't the curent dir, set the out_dir to the parent
         self["out_dir"] = out_file.parent
-        self.logger.info("Resolved out_dir to: %s" % colorize(self["out_dir"], "green"))
+        self.logger.info("Resolved out_dir to: %s" % c_(self["out_dir"], "green"))
         out_file = out_file.name
 
     self.data["out_file"] = out_file
@@ -300,6 +436,18 @@ def _process_paths_multi(self, path: Union[Path, str]) -> None:
     self["paths"].append(path)
 
 
+def _process_libraries_multi(self, library: Union[str]) -> None:
+    """Prociesses libraries into the libraries list, adding the parent directory to the library paths."""
+    if library in self["libraries"]:
+        return self.logger.debug("Library already in libraries list, skipping: %s" % library)
+
+    self.logger.debug("Processing library: %s" % library)
+    library_path = find_library(self, library)
+    self["libraries"].append(library)
+    self["dependencies"] = library_path
+    self["library_paths"] = str(library_path.parent)
+
+
 def _process_binaries_multi(self, binary: str) -> None:
     """Processes binaries into the binaries list, adding dependencies along the way."""
     if binary in self["binaries"]:
@@ -326,10 +474,14 @@ def _process_binaries_multi(self, binary: str) -> None:
 
     self.logger.debug("Adding binary: %s" % binary)
     self["binaries"].append(binary)
+    self["binary_search_paths"] = str(dependencies[0].parent)  # Add the binary path to the search paths
 
 
-def _validate_dependency(self, dependency: Union[Path, str]) -> None:
-    """Performas basic validation and normalization for dependencies."""
+def _validate_dependency(self, dependency: Union[Path, str]) -> Path:
+    """Performas basic validation and normalization for dependencies.
+    Returns the dependency as a Path object.
+    Raises a ValidationError if the dependency path does not exist.
+    """
     if not isinstance(dependency, Path):
         dependency = Path(dependency)
 
@@ -401,15 +553,36 @@ def _process_gz_dependencies_multi(self, dependency: Union[Path, str]) -> None:
     self["gz_dependencies"].append(dependency)
 
 
+def _process_zstd_dependencies_multi(self, dependency: Union[Path, str]) -> None:
+    """Processes zstd dependencies.
+    Checks that the file is a zst file, and adds it to the zstd dependencies list.
+    Adds the dependency to 'dependencies' if the zstandard library is not found.
+    !! Resolves symlinks implicitly !!
+    """
+    if not _has_zstd():
+        self["dependencies"] = dependency
+        return self.logger.warning(
+            f"Python zstandard library not found, adding to dependencies: {c_(dependency, 'yellow', bold=True)}"
+        )
+
+    dependency = _validate_dependency(self, dependency)
+    if dependency.suffix != ".zst":
+        self.logger.warning("ZSTD dependency missing zst extension: %s" % dependency)
+    self["zstd_dependencies"].append(dependency)
+
+
 def _process_build_logging(self, log_build: bool) -> None:
     """Sets the build log flag."""
     build_log_level = self.get("_build_log_level", 10)
+    if log_build == self["build_logging"]:
+        # Don't re-run the setup procedure as it will bump the log level again when args are re-processed
+        return self.logger.debug("Build logging is already set to: %s" % log_build)
     if log_build:
         self["_build_log_level"] = max(build_log_level + 10, 20)
     else:
         if self["_build_log_level"] > 10:
             self.logger.warning("Resetting _build_log_level to 10, as build logging is disabled.")
-        self["_build_log_level"] = 10
+            self["_build_log_level"] = 10
     self.data["build_logging"] = log_build
 
 
@@ -461,7 +634,7 @@ def _process_masks_multi(self, runlevel: str, function: str) -> None:
     if runlevel not in self["masks"]:
         self.logger.debug("Creating new mask: %s" % runlevel)
         self["masks"][runlevel] = NoDupFlatList(logger=self.logger)
-    self.logger.info("[%s] Adding mask: %s" % (runlevel, colorize(function, "red")))
+    self.logger.info("[%s] Adding mask: %s" % (runlevel, c_(function, "red")))
     self["masks"][runlevel] = function
 
 
