@@ -1,10 +1,14 @@
 __author__ = "desultory"
-__version__ = "4.0.3"
+__version__ = "4.1.3"
 
+from json import loads
 from pathlib import Path
+from re import search
+from textwrap import dedent
 
-from ugrd import AutodetectError, ValidationError
-from zenlib.util import colorize, contains, unset
+from ugrd.exceptions import AutodetectError, ValidationError
+from zenlib.util import colorize as c_
+from zenlib.util import contains, unset
 
 _module_name = "ugrd.crypto.cryptsetup"
 
@@ -62,7 +66,7 @@ def _validate_crypysetup_key(self, key_parameters: dict) -> None:
     if key_parameters.get("include_key"):
         return self.logger.info("Skipping key validation for included key.")
     elif key_parameters.get("validate_key") is False:
-        return self.logger.info("Skipping key validation for: %s" % colorize(key_parameters["key_file"], "yellow"))
+        return self.logger.info("Skipping key validation for: %s" % c_(key_parameters["key_file"], "yellow"))
 
     key_path = Path(key_parameters["key_file"])
 
@@ -93,9 +97,7 @@ def _validate_cryptsetup_config(self, mapped_name: str) -> None:
         if not Path(
             config["header_file"]
         ).exists():  # Make sure the header file exists, it may not be present at build time
-            self.logger.warning(
-                "[%s] Header file not found: %s" % (mapped_name, colorize(config["header_file"], "yellow"))
-            )
+            self.logger.warning("[%s] Header file not found: %s" % (mapped_name, c_(config["header_file"], "yellow")))
     elif not any([config.get("partuuid"), config.get("uuid"), config.get("path")]):
         if not self["autodetect_root_luks"]:
             raise ValidationError(
@@ -110,7 +112,7 @@ def _process_cryptsetup_multi(self, mapped_name: str, config: dict) -> None:
     """Processes the cryptsetup configuration"""
     for parameter in config:
         if parameter not in CRYPTSETUP_PARAMETERS:
-            self.logger.error("[%s] Unknown parameter: %s" % (mapped_name, colorize(parameter, "red", bold=True)))
+            self.logger.error("[%s] Unknown parameter: %s" % (mapped_name, c_(parameter, "red", bold=True)))
 
     config = _merge_cryptsetup(self, mapped_name, config)  # Merge the config with the existing configuration
     self.logger.debug("[%s] Processing cryptsetup configuration: %s" % (mapped_name, config))
@@ -161,8 +163,6 @@ def _get_dm_slave_info(self, device_info: dict) -> (str, dict):
 
 def _read_cryptsetup_header(self, mapped_name: str, slave_device: str = None) -> dict:
     """Reads LUKS header information from a device or header file into a dict"""
-    from json import loads
-
     header_file = self["cryptsetup"][mapped_name].get("header_file")
     if not header_file:
         if slave_device:
@@ -186,10 +186,9 @@ def _read_cryptsetup_header(self, mapped_name: str, slave_device: str = None) ->
         return luks_info
     except RuntimeError as e:
         if not self["cryptsetup"][mapped_name].get("header_file"):
-            self.logger.error("Unable to read LUKS header: %s" % colorize(e, "red", bold=True, bright=True))
+            raise AutodetectError(f"Unable to read LUKS header for: {mapped_name}") from e
         else:
-            self.logger.warning("Cannot read detached LUKS header for validation: %s" % colorize(e, "yellow"))
-    return {}
+            raise ValidationError(f"[{mapped_name}] Unable to read detached LUKS header: {header_file}") from e
 
 
 def _detect_luks_aes_module(self, luks_cipher_name: str) -> None:
@@ -203,7 +202,7 @@ def _detect_luks_aes_module(self, luks_cipher_name: str) -> None:
         self.logger.debug("Cipher kernel modules are builtin: %s" % crypto_name)
     else:
         self.logger.info(
-            "[%s] Adding kernel module for LUKS cipher: %s" % (crypto_name, colorize(crypto_config["module"], "cyan"))
+            "[%s] Adding kernel module for LUKS cipher: %s" % (crypto_name, c_(crypto_config["module"], "cyan"))
         )
         self["_kmod_auto"] = crypto_config["module"]
 
@@ -235,11 +234,28 @@ def _validate_cryptsetup_header(self, mapped_name: str) -> None:
     """Validates configured cryptsetup volumes against the LUKS header."""
     cryptsetup_info = self["cryptsetup"][mapped_name]
     if cryptsetup_info.get("validate_header") is False:
-        return self.logger.warning("Skipping cryptsetup header validation for: %s" % colorize(mapped_name, "yellow"))
+        return self.logger.warning("Skipping cryptsetup header validation for: %s" % c_(mapped_name, "yellow"))
 
-    luks_info = _read_cryptsetup_header(self, mapped_name)
-    if not luks_info:
-        raise ValidationError("[%s] Unable to read LUKS header." % mapped_name)
+    try:
+        luks_info = _read_cryptsetup_header(self, mapped_name)
+    except (AutodetectError, ValidationError) as e:
+        self.logger.warning(f"Unable to validate LUKS header for: {c_(mapped_name, 'red', bold=True)}\n")
+        if "header_file" not in cryptsetup_info:
+            self.logger.warning("If the header is detached, please set `header_file` in the configuration.")
+        self.logger.warning(
+            "If the header is detached not accessible at build time, please set `validate_header = false`."
+        )
+        self.logger.critical(
+            "When header validation is disabled, it's up to the user to ensure valid headers are accessible at boot time, and that cryptsetup is built with the correct dependencies!"
+        )
+        self.logger.warning(
+            "Validation is not supported for LUKS1. If LUKS1 is being used, please consider using LUKS2 if possible.\n"
+        )
+        self.logger.info(
+            f"Header validation can be disabled for this LUKS volume by setting the following configuration:\n\n[cryptsetup.{mapped_name}]\nvalidate_header = false\n\nor by setting `cryptsetup_header_validation = false` globally.\n"
+        )
+
+        raise e
 
     if uuid := cryptsetup_info.get("uuid"):
         if luks_info.get("uuid") != uuid:
@@ -271,7 +287,7 @@ def _validate_cryptsetup_device(self, mapped_name) -> None:
 
     cryptsetup_info = self["cryptsetup"][mapped_name]  # Get the cryptsetup information
     if cryptsetup_info.get("validate") is False:
-        return self.logger.warning("Skipping cryptsetup device validation: %s" % colorize(mapped_name, "yellow"))
+        return self.logger.warning("Skipping cryptsetup device validation: %s" % c_(mapped_name, "yellow"))
 
     slave_device, blkid_info = _get_dm_slave_info(self, dm_info)  # Get the blkid information
 
@@ -292,8 +308,6 @@ def _validate_cryptsetup_device(self, mapped_name) -> None:
 @unset("_cryptsetup_backend")
 def detect_cryptsetup_backend(self) -> None:
     """Determines the cryptsetup backend by running 'cryptsetup --debug luksDump' on this file"""
-    from re import search
-
     try:
         raw_luks_info = (
             self._run(["cryptsetup", "--debug", "luksDump", __file__], fail_silent=True, fail_hard=False)
@@ -304,9 +318,7 @@ def detect_cryptsetup_backend(self) -> None:
             if line.startswith("# Crypto backend"):
                 backend = search(r"backend \((.+)\)", line).group(1)
                 self["_cryptsetup_backend"] = backend.split()[0].lower()
-                return self.logger.info(
-                    "Detected cryptsetup backend: %s" % colorize(self["_cryptsetup_backend"], "cyan")
-                )
+                return self.logger.info("Detected cryptsetup backend: %s" % c_(self["_cryptsetup_backend"], "cyan"))
     except RuntimeError as e:
         self.logger.error("Unable to determine cryptsetup backend: %s" % e)
 
@@ -392,15 +404,19 @@ def _validate_luks_config(self, mapped_name: str) -> None:
 
 def export_crypt_sources(self) -> list[str]:
     """Validates the cryptsetup configuration (if enabled).
+    Sets CRYPTSETUP_HEADER_{name} if a header file is set.
     Adds the cryptsetup source and token to the exports.
     Sets the token to the partuuid or uuid if it exists.
     Sets the SOURCE when using a path.
     Only allows using the path if validation is disabled.
-    sets CRYPTSETUP_HEADER_{name} if a header file is set.
     """
     for name, parameters in self["cryptsetup"].items():
-        _validate_luks_config(self, name)
-        if parameters.get("path"):
+        _validate_luks_config(self, name)  # First validate the configuration
+        if header_file := parameters.get("header_file"):  # Then add the header file to the exports, if defined
+            self["exports"]["CRYPTSETUP_HEADER_%s" % name] = header_file
+            self.logger.debug("Set CRYPTSETUP_HEADER_%s: %s" % (name, header_file))
+
+        if parameters.get("path"):  # If a path is set, only allow it if validation is disabled
             if not self["validate"]:
                 self.logger.warning(
                     "Using device paths is unreliable and can result in boot failures. Consider using partuuid."
@@ -412,16 +428,13 @@ def export_crypt_sources(self) -> list[str]:
         elif not parameters.get("partuuid") and not parameters.get("uuid") and parameters.get("path"):
             raise ValidationError("Device source for cryptsetup mount must be specified: %s" % name)
 
-        for token_type in ["partuuid", "uuid"]:
+        for token_type in ["partuuid", "uuid"]:  # Attempt to set the token from partuuid or uuid
             if token := parameters.get(token_type):
                 self["exports"]["CRYPTSETUP_TOKEN_%s" % name] = f"{token_type.upper()}={token}"
                 self.logger.debug("Set CRYPTSETUP_TOKEN_%s: %s=%s" % (name, token_type.upper(), token))
                 break
-        else:
+        else:  # Raise an error if no usable token/config is set
             raise ValidationError("A partuuid or uuid must be specified for cryptsetup mount: %s" % name)
-        if header_file := parameters.get("header_file"):
-            self["exports"]["CRYPTSETUP_HEADER_%s" % name] = header_file
-            self.logger.debug("Set CRYPTSETUP_HEADER_%s: %s" % (name, header_file))
 
 
 def get_crypt_dev(self) -> str:
@@ -444,9 +457,6 @@ def get_crypt_dev(self) -> str:
             printf "%s" "$source_dev"
         fi
     fi
-    if [ -z "$source_dev" ]; then
-        rd_fail "Failed to find cryptsetup device: $1"
-    fi
     """
 
 
@@ -458,6 +468,10 @@ def open_crypt_dev(self) -> str:
     """
     out = """
     crypt_device="$(get_crypt_dev "$1")"
+    if [ -z "$crypt_device" ]; then
+        rd_fail "Failed to find cryptsetup device: $1"
+    fi
+
     header="$(readvar CRYPTSETUP_HEADER_"$1")"
 
     cryptsetup_args="cryptsetup open --tries 1"
@@ -465,9 +479,9 @@ def open_crypt_dev(self) -> str:
         einfo "[$1] Opening cryptsetup device with detached header: $header"
         cryptsetup_args="$cryptsetup_args --header $header"
     fi
-    if [ -e /run/vars/key_data ]; then
-        einfo "[$1] Unlocking with key data from: /run/vars/key_data"
-        cryptsetup_args="$cryptsetup_args --key-file /run/vars/key_data"
+    if [ -e /run/ugrd/key_data ]; then
+        einfo "[$1] Unlocking with key data from: /run/ugrd/key_data"
+        cryptsetup_args="$cryptsetup_args --key-file /run/ugrd/key_data"
     elif [ -n "$2" ]; then
         einfo "[$1] Unlocking with key file: $2"
         cryptsetup_args="$cryptsetup_args --key-file $2"
@@ -487,8 +501,6 @@ def open_crypt_dev(self) -> str:
 
 def _open_crypt_dev(self, name: str, parameters: dict) -> list[str]:
     """Generates a loop to open a cryptsetup device with the given parameters."""
-    from textwrap import dedent
-
     retries = parameters.get("retries", self["cryptsetup_retries"])
     out = [
         f"einfo 'Opening cryptsetup device: {name}'",
@@ -501,21 +513,27 @@ def _open_crypt_dev(self, name: str, parameters: dict) -> list[str]:
     key_command = parameters.get("key_command")
     plymouth_key_command = parameters.get("plymouth_key_command") if "ugrd.base.plymouth" in self["modules"] else None
 
-    _key_command_lines = f"""    einfo "($i/$retries)[{name}] Running key command: {key_command}"
-    if ! {key_command} > /run/vars/key_data; then
-        ewarn 'Failed to run key command: {key_command}'
-        {reset_command}
-        continue
-    fi"""
+    # Build the key command lines
+    _key_command_lines = "\n".join(
+        [
+            f'einfo "($i/$retries)[{name}] Running key command: {key_command}"',
+            f"if ! {key_command} > /run/ugrd/key_data; then",
+            f"    ewarn 'Failed to run key command: {key_command}'",
+            "    prompt_user 'Press space to retry'" if not self["cryptsetup_autoretry"] else "",
+            f"    {reset_command}" if reset_command else "",
+            "    continue",
+            "fi",
+        ]
+    )
 
     if key_command:
         self.logger.debug("[%s] Using key command: %s" % (name, key_command))
-        out += ["    rm -f /run/vars/key_data"]  # Remove the key data file if it exists
+        out += ["    rm -f /run/ugrd/key_data"]  # Remove the key data file if it exists
         if plymouth_key_command:
             self.logger.debug("[%s] Using plymouth key command: %s" % (name, plymouth_key_command))
             out += [
                 "    if plymouth --ping; then",
-                f'        if ! plymouth ask-for-password --prompt "[${{i}} / ${{retries}}] Enter passphrase to unlock key for: {name}" --command "{parameters["plymouth_key_command"]}" --number-of-tries 1 > /run/vars/key_data; then',
+                f'        if ! plymouth ask-for-password --prompt "[${{i}} / ${{retries}}] Enter passphrase to unlock key for: {name}" --command "{parameters["plymouth_key_command"]}" --number-of-tries 1 > /run/ugrd/key_data; then',
                 *[f"            {line}" for line in reset_lines],
                 "        fi",
                 "    else",
@@ -527,7 +545,7 @@ def _open_crypt_dev(self, name: str, parameters: dict) -> list[str]:
         out += [  # open_crypt_dev will use the key data file if it exists
             f'    einfo "($i/$retries)[{name}] Unlocked key to var: key_data"',
             f"    if open_crypt_dev {name}; then",
-            "        rm -f /run/vars/key_data",  # Remove the key data file after opening the device
+            "        rm -f /run/ugrd/key_data",  # Remove the key data file after opening the device
         ]
     elif key_file:  # For plain key files, just use it as arg 2 for open_crypt_dev
         out += [
@@ -538,7 +556,7 @@ def _open_crypt_dev(self, name: str, parameters: dict) -> list[str]:
         if "ugrd.base.plymouth" in self["modules"]:
             out += [  # When plymouth is used, write the entered key to the key data file, where it will be used by open_crypt_dev
                 "    if plymouth --ping; then",
-                f'        plymouth ask-for-password --prompt "[${{i}} / ${{retries}}] Enter passphrase to unlock: {name}" > /run/vars/key_data',
+                f'        plymouth ask-for-password --prompt "[${{i}} / ${{retries}}] Enter passphrase to unlock: {name}" > /run/ugrd/key_data',
                 "    fi",
                 f'    einfo "($i/$retries) Unlocking device: {name}"',
                 f"    if open_crypt_dev {name}; then",
@@ -548,15 +566,15 @@ def _open_crypt_dev(self, name: str, parameters: dict) -> list[str]:
     # Break on success
     out += ["        break", "    fi", f'    ewarn "($i/$retries) Failed to open cryptsetup device: {name}"']
     if not self["cryptsetup_autoretry"]:
-        out += ["    prompt_user 'Press enter to retry'"]
+        out += ["    prompt_user 'Press space to retry'"]
     if reset_command:
         out += ['    einfo "Running key reset command"', f"    {reset_command}"]
 
     out += [
         "done",
-        "if [ -e /run/vars/key_data ]; then",
+        "if [ -e /run/ugrd/key_data ]; then",
         "    eerror 'Removing leftover key data file'",
-        "    rm -f /run/vars/key_data",
+        "    rm -f /run/ugrd/key_data",
         "fi",
     ]
 
