@@ -1,11 +1,17 @@
 from pathlib import Path
+from shutil import copy2
 from subprocess import CompletedProcess, TimeoutExpired, run
 from typing import Union
+from uuid import uuid4
 
-from zenlib.util import pretty_print, colorize
+from zenlib.util import colorize as c_
+from zenlib.util import pretty_print
 
-__version__ = "1.5.5"
-__author__ = "desultory"
+from .exceptions import ValidationError
+
+__version__ = "1.7.1"
+
+_RANDOM_BUILD_ID = str(uuid4())
 
 
 def get_subpath(path: Path, subpath: Union[Path, str]) -> Path:
@@ -32,8 +38,13 @@ class GeneratorHelpers:
         return get_subpath(get_subpath(self.tmpdir, self.out_dir), path)
 
     def _get_build_path(self, path: Union[Path, str]) -> Path:
-        """Returns the path relative to the build directory, under the tmpdir."""
-        return get_subpath(get_subpath(self.tmpdir, self.build_dir), path)
+        """Returns the path relative to the build directory, under the tmpdir.
+        if random_build_dir is true, appends a uuid4() to the build directory."""
+        if self.random_build_dir:
+            build_dir = self.build_dir.with_name(self.build_dir.name + "-" + _RANDOM_BUILD_ID)
+        else:
+            build_dir = self.build_dir
+        return get_subpath(get_subpath(self.tmpdir, build_dir), path)
 
     def _mkdir(self, path: Path, resolve_build=True) -> None:
         """
@@ -61,45 +72,51 @@ class GeneratorHelpers:
 
         if not path_dir.is_dir():
             path_dir.mkdir()
-            self.logger.log(self["_build_log_level"], "Created directory: %s" % path)
+            self.logger.log(self["_build_log_level"], "Created directory: %s" % c_(path, "green"))
         else:
             self.logger.debug("Directory already exists: %s" % path_dir)
 
-    def _write(self, file_name: Union[Path, str], contents: list[str], chmod_mask=0o644) -> None:
+    def _write(self, file_name: Union[Path, str], contents: list[str], chmod_mask=0o644, append=False) -> None:
         """
         Writes a file within the build directory.
         Sets the passed chmod_mask.
         If the first line is a shebang, sh -n is run on the file.
         """
-        from os import chmod
-
         file_path = self._get_build_path(file_name)
 
         if not file_path.parent.is_dir():
             self.logger.debug("Parent directory for '%s' does not exist: %s" % (file_path.name, file_path))
             self._mkdir(file_path.parent, resolve_build=False)
 
+        if isinstance(contents, list):
+            contents = "\n".join(contents)
+
         if file_path.is_file():
-            self.logger.warning("File already exists: %s" % colorize(file_path, "yellow"))
-            if self.clean:
-                self.logger.warning("Deleting file: %s" % colorize(file_path, "red", bright=True, bold=True))
+            self.logger.warning("File already exists: %s" % c_(file_path, "yellow"))
+            if contents in file_path.read_text():
+                self.logger.debug("Contents:\n%s" % contents)
+                return self.logger.warning("Contents are already present, skipping write: %s" % file_path)
+
+            if self.clean and not append:
+                self.logger.warning("Deleting file: %s" % c_(file_path, "red", bright=True, bold=True))
                 file_path.unlink()
 
         self.logger.debug("[%s] Writing contents:\n%s" % (file_path, contents))
-        with open(file_path, "w") as file:
-            file.writelines("\n".join(contents))
+        with open(file_path, "a") as file:
+            file.write(contents)
 
-        if contents[0].startswith(self["shebang"].split(" ")[0]):
+        if contents.startswith(self["shebang"].split(" ")[0]):
             self.logger.debug("Running sh -n on file: %s" % file_name)
             try:
                 self._run(["sh", "-n", str(file_path)])
             except RuntimeError as e:
-                raise RuntimeError("Failed to validate shell script: %s" % pretty_print(contents)) from e
-        elif contents[0].startswith("#!"):
+                self.logger.error(f"Invalid shell script:\n{pretty_print(contents)}")
+                raise ValidationError(f"Failed to validate shell script: {file_name}") from e
+        elif contents.startswith("#!"):
             self.logger.warning("[%s] Skipping sh -n on file with unrecognized shebang: %s" % (file_name, contents[0]))
 
-        self.logger.info("Wrote file: %s" % colorize(file_path, "green", bright=True))
-        chmod(file_path, chmod_mask)
+        self.logger.info("Wrote file: %s" % c_(file_path, "green", bright=True))
+        file_path.chmod(chmod_mask)
         self.logger.debug("[%s] Set file permissions: %s" % (file_path, chmod_mask))
 
     def _copy(self, source: Union[Path, str], dest=None) -> None:
@@ -111,8 +128,6 @@ class GeneratorHelpers:
 
         Raises a RuntimeError if the destination path is not within the build directory.
         """
-        from shutil import copy2
-
         if not isinstance(source, Path):
             source = Path(source)
 
@@ -133,7 +148,7 @@ class GeneratorHelpers:
             self._mkdir(dest_path.parent, resolve_build=False)
 
         if dest_path.is_file():
-            self.logger.warning("File already exists, overwriting: %s" % colorize(dest_path, "yellow", bright=True))
+            self.logger.warning("File already exists, overwriting: %s" % c_(dest_path, "yellow", bright=True))
         elif dest_path.is_dir():
             self.logger.debug("Destination is a directory, adding source filename: %s" % source.name)
             dest_path = dest_path / source.name
@@ -143,7 +158,7 @@ class GeneratorHelpers:
         except ValueError as e:
             raise RuntimeError("Destination path is not within the build directory: %s" % dest_path) from e
 
-        self.logger.log(self["_build_log_level"], "Copying '%s' to '%s'" % (source, dest_path))
+        self.logger.log(self["_build_log_level"], "Copying '%s' to '%s'" % (c_(source, "blue"), c_(dest_path, "green")))
         copy2(source, dest_path)
 
     def _symlink(self, source: Union[Path, str], target: Union[Path, str]) -> None:
@@ -179,7 +194,7 @@ class GeneratorHelpers:
             if target.resolve() == source:
                 return self.logger.debug("Symlink already exists: %s -> %s" % (target, source))
             elif self.clean:
-                self.logger.warning("Deleting symlink: %s" % colorize(target, "red", bright=True))
+                self.logger.warning("Deleting symlink: %s" % c_(target, "red", bright=True))
                 target.unlink()
             else:
                 raise RuntimeError("Symlink already exists: %s -> %s" % (target, target.resolve()))
@@ -187,26 +202,43 @@ class GeneratorHelpers:
         if target.relative_to(self._get_build_path("/")) == source:
             return self.logger.debug("Cannot symlink to self: %s -> %s" % (target, source))
 
-        self.logger.debug("Creating symlink: %s -> %s" % (target, source))
+        self.logger.log(
+            self["_build_log_level"], "Creating symlink: %s -> %s" % (c_(target, "green"), c_(source, "blue"))
+        )
         target.symlink_to(source)
 
     def _run(self, args: list[str], timeout=None, fail_silent=False, fail_hard=True) -> CompletedProcess:
-        """Runs a command, returns the CompletedProcess object"""
+        """Runs a command, returns the CompletedProcess object on success.
+        If a timeout is set, the command will fail hard if it times out.
+        If fail_silent is set, non-zero return codes will not log stderr/stdout.
+        If fail_hard is set, non-zero return codes will raise a RuntimeError.
+        """
+
+        def print_err(ret) -> None:
+            if args := ret.args:
+                if isinstance(args, tuple):
+                    args = args[0]  # When there's a timeout, args is a (args, timeout) tuple
+                self.logger.error("Failed command: %s" % c_(" ".join(args), "red", bright=True))
+            if stdout := ret.stdout:
+                self.logger.error("Command output:\n%s" % stdout.decode())
+            if stderr := ret.stderr:
+                self.logger.error("Command error:\n%s" % stderr.decode())
+
         timeout = timeout or self.timeout
         cmd_args = [str(arg) for arg in args]
         self.logger.debug("Running command: %s" % " ".join(cmd_args))
         try:
             cmd = run(cmd_args, capture_output=True, timeout=timeout)
         except TimeoutExpired as e:
+            # Always fail hard for timeouts
+            print_err(e)
             raise RuntimeError("[%ds] Command timed out: %s" % (timeout, [str(arg) for arg in cmd_args])) from e
 
         if cmd.returncode != 0:
             if not fail_silent:
-                self.logger.error("Failed to run command: %s" % colorize(" ".join(cmd.args), "red", bright=True))
-                self.logger.error("Command output:\n%s" % cmd.stdout.decode())
-                self.logger.error("Command error:\n%s" % cmd.stderr.decode())
-            if fail_hard:
-                raise RuntimeError("Failed to run command: %s" % " ".join(cmd.args))
+                print_err(cmd)  # Print the full error output if not failing silently
+            if fail_hard:  # Fail hard means raise an exception
+                raise RuntimeError("Failed to run command: %s" % c_(" ".join(cmd.args), "red", bright=True))
 
         return cmd
 
@@ -220,7 +252,7 @@ class GeneratorHelpers:
         # If the cycle count is not set, attempt to clean
         if not self.old_count:
             if self.clean:
-                self.logger.warning("Deleting file: %s" % colorize(file_name, "red", bold=True, bright=True))
+                self.logger.warning("Deleting file: %s" % c_(file_name, "red", bold=True, bright=True))
                 file_name.unlink()
                 return
             else:
@@ -242,7 +274,7 @@ class GeneratorHelpers:
             if sequence >= self.old_count:
                 # Clean the last file in the sequence if clean is enabled
                 if self.clean:
-                    self.logger.warning("Deleting old file: %s" % colorize(target_file, "red", bold=True, bright=True))
+                    self.logger.warning("Deleting old file: %s" % c_(target_file, "red", bold=True, bright=True))
                     target_file.unlink()
                 else:
                     self.logger.debug("Cycle limit reached")
@@ -254,3 +286,86 @@ class GeneratorHelpers:
         # Finally, rename the file
         self.logger.info("[%d] Cycling file: %s -> %s" % (sequence, file_name, target_file))
         file_name.rename(target_file)
+
+    def sort_hook_functions(self, hook: str) -> None:
+        """Sorts the functions for the specified hook based on the import order.
+        "before" functions are moved before the target function,
+        "after" functions' target function is moved before the current function.
+
+        Filters orders which do not contain functions or targets in the current hook.
+        """
+        func_names = [func.__name__ for func in self["imports"].get(hook, [])]
+        if not func_names:
+            return self.logger.debug("No functions for hook: %s" % hook)
+
+        b = self["import_order"].get("before", {})
+        before = {k: v for k, v in b.items() if k in func_names and any(subv in func_names for subv in b[k])}
+        a = self["import_order"].get("after", {})
+        after = {k: v for k, v in a.items() if k in func_names and any(subv in func_names for subv in a[k])}
+
+        if not before and not after:
+            return self.logger.debug("No import order specified for hook: %s" % hook)
+
+        def iter_order(order, direction):
+            """Iterate over all functions in an import order list,
+            using this information to move the order of function names in the import list.
+
+            Returns True if any changes were made, False otherwise."""
+            changed = False
+
+            for func_name, other_funcs in order.items():
+                func_index = func_names.index(func_name)  # Get the index based on the current position
+                assert func_index >= 0, "Function not found in import list: %s" % func_name
+                for other_func in other_funcs:
+                    try:
+                        other_index = func_names.index(other_func)
+                    except ValueError:
+                        continue
+                    assert other_index >= 0, "Function not found in import list: %s" % other_func
+
+                    def reorder_func(direction):
+                        """Reorders the function based on the direction."""
+                        if direction == "before":  # Move the function before the other function
+                            self.logger.debug("[%s] Moving function before: %s" % (func_name, other_func))
+                            func_names.insert(other_index, func_names.pop(func_index))
+                            self["imports"][hook].insert(other_index, self["imports"][hook].pop(func_index))
+                        elif direction == "after":  # Move the other function before the current function
+                            self.logger.debug("[%s] Moving function before: %s" % (other_func, func_name))
+                            func_names.insert(func_index, func_names.pop(other_index))
+                            self["imports"][hook].insert(func_index, self["imports"][hook].pop(other_index))
+                        else:
+                            raise ValueError("Invalid direction: %s" % direction)
+
+                    self.logger.log(5, "[%s] Imports:\n%s", hook, ", ".join(i.__name__ for i in self["imports"][hook]))
+                    if direction == "before":  # func_index should be before other_index
+                        if func_index > other_index:  # If the current function is after the other function
+                            reorder_func("before")  # Move the current function before the other function)
+                            changed = True
+                        else:  # Log otherwise
+                            self.logger.log(5, "Function %s already before: %s" % (func_name, other_func))
+                    elif direction == "after":  # func_index should be after other_index
+                        if func_index < other_index:  # If the current function is before the other function
+                            reorder_func("after")  # Move the current function after the other function
+                            changed = True
+                        else:
+                            self.logger.log(5, "Function %s already after: %s" % (func_name, other_func))
+                    else:
+                        raise ValueError("Invalid direction: %s" % direction)
+                    func_index = func_names.index(func_name)  # Update the index after moving
+            return changed
+
+        max_iterations = len(func_names) * (len(before) + 1) * (len(after) + 1)  # Prevent infinite loops
+        iterations = max_iterations
+        while iterations:
+            iterations -= 1
+            # Update the order based on the before and after lists
+            if not any([iter_order(before, "before"), iter_order(after, "after")]):
+                self.logger.debug(
+                    "[%s] Import order converged after %s iterations" % (hook, max_iterations - iterations)
+                )
+                break  # Keep going until no changes are made
+        else:  # If the loop completes without breaking, raise an error
+            self.logger.error("Import list: %s" % func_names)
+            self.logger.error("Before: %s" % before)
+            self.logger.error("After: %s" % after)
+            raise ValueError("Import order did not converge after %s iterations" % max_iterations)
