@@ -1,10 +1,14 @@
+from importlib.metadata import version
+from textwrap import dedent
 from tomllib import TOMLDecodeError, load
 
 from zenlib.logging import loggify
-from zenlib.util import colorize, pretty_print
+from zenlib.util import colorize as c_
+from zenlib.util import pretty_print
 
 from ugrd.initramfs_dict import InitramfsConfigDict
 
+from .exceptions import ValidationError
 from .generator_helpers import GeneratorHelpers
 
 
@@ -20,28 +24,19 @@ class InitramfsGenerator(GeneratorHelpers):
         self.build_tasks = ["build_enum", "build_pre", "build_tasks", "build_late", "build_deploy", "build_final"]
 
         # init_pre and init_final are run as part of generate_initramfs_main
-        self.init_types = [
-            "init_debug",
-            "init_early",
-            "init_main",
-            "init_late",
-            "init_premount",
-            "init_mount",
-            "init_mount_late",
-            "init_cleanup",
-        ]
+        self.init_types = ["init_debug", "init_main", "init_mount"]
 
         # Passed kwargs must be imported early, so they will be processed against the base configuration
         self.config_dict.import_args(kwargs)
-        try:
+        try:  # Attempt to load the config file, if it exists
             self.load_config(config)  # The user config is loaded over the base config, clobbering kwargs
             self.config_dict.import_args(
                 kwargs, quiet=True
             )  # Re-import kwargs (cmdline params) to apply them over the config
         except FileNotFoundError:
-            if config:
-                self.logger.warning("[%s] Config file not found, using the base config." % config)
-            else:
+            if config:  # If a config file was specified, log an error that it's missing
+                self.logger.critical("[%s] Config file not found, using the base config." % config)
+            else:  # Otherwise, log info that the base config is being used
                 self.logger.info("No config file specified, using the base config.")
         except TOMLDecodeError as e:
             raise ValueError("[%s] Error decoding config file: %s" % (config, e))
@@ -56,7 +51,7 @@ class InitramfsGenerator(GeneratorHelpers):
             raise FileNotFoundError("Config file not specified.")
 
         with open(config_filename, "rb") as config_file:
-            self.logger.info("Loading config file: %s" % colorize(config_file.name, "blue", bold=True, bright=True))
+            self.logger.info("Loading config file: %s" % c_(config_file.name, "blue", bold=True, bright=True))
             raw_config = load(config_file)
 
         # Process into the config dict, it should handle parsing
@@ -69,6 +64,7 @@ class InitramfsGenerator(GeneratorHelpers):
 
         self.logger.debug("Loaded config:\n%s" % self.config_dict)
 
+    #  If the initramfs generator is used as a dictionary, it will use the config_dict.
     def __setitem__(self, key, value):
         self.config_dict[key] = value
 
@@ -89,12 +85,15 @@ class InitramfsGenerator(GeneratorHelpers):
 
     def build(self) -> None:
         """Builds the initramfs image."""
-        from importlib.metadata import version
-
         self._log_run(f"Running ugrd v{version('ugrd')}")
         self.run_build()
         self.config_dict.validate()  # Validate the config after the build tasks have been run
 
+        if self.validate and not self.validated:
+            raise ValidationError(
+                f"Failed to validate config. Unprocessed values: {', '.join(list(self['_processing'].keys()))}"
+            )
+
         self.generate_init()
         self.pack_build()
         self.run_checks()
@@ -106,7 +105,7 @@ class InitramfsGenerator(GeneratorHelpers):
         If force_include is set, forces the function to be included in the shell profile.
         if force_exclude is set, does not include the output of the function in the shell profile.
         """
-        self.logger.log(self["_build_log_level"], "Running function: %s" % function.__name__)
+        self.logger.log(self["_build_log_level"], "Running function: %s" % c_(function.__name__, "blue", bold=True))
 
         if function_output := function(self):
             if isinstance(function_output, list) and len(function_output) == 1:
@@ -123,7 +122,6 @@ class InitramfsGenerator(GeneratorHelpers):
                 return function_output
 
             if isinstance(function_output, str) and "\n" in function_output:
-                from textwrap import dedent
                 function_output = dedent(function_output)
                 function_output = [  # If the output string has a newline, split and get rid of empty lines
                     line for line in function_output.split("\n") if line and line != "\n" and not line.isspace()
@@ -148,23 +146,25 @@ class InitramfsGenerator(GeneratorHelpers):
             self.logger.debug("[%s] Function returned no output" % function.__name__)
 
     def run_hook(self, hook: str, *args, **kwargs) -> list[str]:
-        """Runs a hook for imported functions."""
+        """Runs all functions for the specified hook.
+        If the function is masked, it will be skipped.
+        If the function is in import_order, handle the ordering
+        """
+        self.sort_hook_functions(hook)  # This is in generator_helpers.py
         out = []
         for function in self["imports"].get(hook, []):
-            # Check that the function is not masked
             if function.__name__ in self["masks"].get(hook, []):
                 self.logger.warning(
-                    "[%s] Skipping masked function: %s" % (hook, colorize(function.__name__, "yellow", bold=True))
+                    f"[{c_(hook, bright=True)}] Skipping masked function: {c_(function.__name__, 'yellow', bold=True)}"
                 )
                 continue
+
             if function_output := self.run_func(function, *args, **kwargs):
                 out.append(function_output)
         return out
 
     def generate_profile(self) -> list[str]:
         """Generates the shell profile file based on self.included_functions."""
-        from importlib.metadata import version
-
         ver = version(__package__) or 9999  # Version won't be found unless the package is installed
         out = [
             self["shebang"].split(" ")[0],  # Don't add arguments to the shebang (for the profile)
@@ -176,6 +176,11 @@ class InitramfsGenerator(GeneratorHelpers):
         self.logger.debug("Library paths: %s" % library_paths)
         out.append(f"export LD_LIBRARY_PATH={library_paths}")
 
+        # Add search paths
+        search_paths = ":".join(self["binary_search_paths"])
+        self.logger.debug("Search paths: %s" % search_paths)
+        out.append(f"export PATH={search_paths}")
+
         for func_name, func_content in self.included_functions.items():
             out.append("\n\n" + func_name + "() {")
             if isinstance(func_content, str):
@@ -213,7 +218,7 @@ class InitramfsGenerator(GeneratorHelpers):
         init.extend(self.run_init_hook("init_pre"))  # Always run init_pre first
 
         # If custom_init is used, create the init using that
-        if self["imports"].get("custom_init") and self.get("_custom_init_file"):
+        if self["imports"].get("custom_init"):
             init += ["\n# !!custom_init"]
             init_line, custom_init = self["imports"]["custom_init"](self)
             if isinstance(init_line, str):
@@ -221,6 +226,7 @@ class InitramfsGenerator(GeneratorHelpers):
             else:
                 init.extend(init_line)
         else:  # Otherwise, use the standard init generator
+            custom_init = None
             init.extend(self.generate_init_main())
 
         init.extend(self.run_init_hook("init_final"))  # Always run init_final last
@@ -230,7 +236,8 @@ class InitramfsGenerator(GeneratorHelpers):
             self._write("/etc/profile", self.generate_profile(), 0o755)
             self.logger.info("Included functions: %s" % ", ".join(list(self.included_functions.keys())))
 
-        if self.get("_custom_init_file"):  # Write the custom init file if it exists
+        # Write the custom init file if it exists
+        if custom_init:
             self._write(self["_custom_init_file"], custom_init, 0o755)
 
         self._write("init", init, 0o755)
@@ -251,7 +258,7 @@ class InitramfsGenerator(GeneratorHelpers):
         else:
             self.logger.warning(
                 "No pack functions specified, the final build is present in: %s"
-                % colorize(self.build_dir, "green", bold=True, bright=True)
+                % c_(self.build_dir, "green", bold=True, bright=True)
             )
 
     def run_init_hook(self, level: str) -> list[str]:
@@ -274,8 +281,6 @@ class InitramfsGenerator(GeneratorHelpers):
             else:
                 self.logger.warning("No checks executed.")
         except (FileNotFoundError, ValueError) as e:
-            from . import ValidationError
-
             raise ValidationError(f"Error running checks: {e}") from e
 
     def run_tests(self) -> None:
@@ -287,7 +292,7 @@ class InitramfsGenerator(GeneratorHelpers):
             self.logger.debug("No tests executed.")
 
     def _log_run(self, logline) -> None:
-        self.logger.info(f"-- | {colorize(logline, 'blue', bold=True)}")
+        self.logger.info(f"-- | {c_(logline, 'blue', bold=True)}")
 
     def __str__(self) -> str:
         return str(self.config_dict)
