@@ -1,11 +1,22 @@
 __author__ = "desultory"
-__version__ = "3.2.2"
+__version__ = "3.4.1"
 
 from pathlib import Path
+from platform import uname
+from struct import error as StructError
+from struct import unpack
 from subprocess import run
 
-from ugrd.kmod import BuiltinModuleError, DependencyResolutionError, IgnoredModuleError, _normalize_kmod_name
-from zenlib.util import colorize, contains, unset
+from ugrd.exceptions import AutodetectError, ValidationError
+from ugrd.kmod import (
+    BuiltinModuleError,
+    DependencyResolutionError,
+    IgnoredModuleError,
+    MissingModuleError,
+    _normalize_kmod_name,
+)
+from zenlib.util import colorize as c_
+from zenlib.util import contains, unset
 
 MODULE_METADATA_FILES = ["modules.order", "modules.builtin", "modules.builtin.modinfo"]
 
@@ -32,6 +43,20 @@ def _process_kmod_init_multi(self, module: str) -> None:
     self["kernel_modules"] = module
 
 
+def _process_kmod_init_optional_multi(self, module: str) -> None:
+    """Adds an optional kmod init module"""
+    module = _normalize_kmod_name(module)
+    if module in self["kmod_ignore"]:
+        self.logger.warning(f"Optional kmod_init module is in the ignore list: {c_(module, 'yellow', bold=True)}")
+        self["_kmod_removed"] = module
+        return
+    if module in self["kmod_init"]:
+        self.logger.debug(f"Optional kmod_init module is already in kmod_init: {c_(module, 'yellow', bold=True)}")
+        return
+    self.logger.debug(f"Adding optional kmod_init module: {c_(module, 'magenta')}")
+    self["kmod_init_optional"].append(module)
+
+
 def _process__kmod_auto_multi(self, module: str) -> None:
     """Adds autodetected modules to self['kernel_modules']."""
     module = _normalize_kmod_name(module)
@@ -91,23 +116,19 @@ def _get_kmod_info(self, module: str):
 
 @contains("kmod_autodetect_lspci", "kmod_autodetect_lspci is not enabled, skipping.")
 def _autodetect_modules_lspci(self) -> None:
-    """Gets the name of all kernel modules being used by hardware visible in lspci -k."""
-    try:
-        cmd = self._run(["lspci", "-k"])
-    except RuntimeError as e:
-        raise DependencyResolutionError("Failed to get list of kernel modules") from e
+    """Uses /sys/bus/pci/drivers to get a list of all kernel modules.
+    Similar to lspci -k."""
     lspci_kmods = set()
-    # Iterate over all output lines
-    for line in cmd.stdout.decode("utf-8").split("\n"):
-        # If the line contains the string 'Kernel modules:' or 'Kernel driver in use:', it contains the name of a kernel module
-        if "Kernel modules:" in line or "Kernel driver in use:" in line:
-            module = line.split(":")[1]
-            if "," in module:
-                # If there are multiple modules, split them and add them to the module set
-                for module in module.split(","):
-                    lspci_kmods.add(module.strip())
-            else:
-                lspci_kmods.add(module.strip())
+    for driver in Path("/sys/bus/pci/drivers").iterdir():
+        if not driver.is_dir():
+            self.logger.debug("Skipping non-directory: %s" % driver)
+            continue
+        module = driver / "module"
+        if not module.exists():
+            self.logger.debug("Skipping driver without module: %s" % driver)
+            continue
+        lspci_kmods.add(module.resolve().name)
+        self.logger.debug("[%s] Autodetected kernel module: %s" % (driver, module.resolve().name))
 
     self["_kmod_auto"] = list(lspci_kmods)
 
@@ -115,8 +136,6 @@ def _autodetect_modules_lspci(self) -> None:
 @contains("kmod_autodetect_lsmod", "kmod_autodetect_lsmod is not enabled, skipping.")
 def _autodetect_modules_lsmod(self) -> None:
     """Gets the name of all currently used kernel modules."""
-    from platform import uname
-
     if self.get("kernel_version") and self["kernel_version"] != uname().release:
         self.logger.warning(
             "Kernel version is set to %s, but the current kernel version is %s"
@@ -138,7 +157,7 @@ def autodetect_modules(self) -> None:
     _autodetect_modules_lsmod(self)
     _autodetect_modules_lspci(self)
     if self["_kmod_auto"]:
-        self.logger.info("Autodetected kernel modules: %s" % colorize(", ".join(self["_kmod_auto"]), "cyan"))
+        self.logger.info("Autodetected kernel modules: %s" % c_(", ".join(self["_kmod_auto"]), "cyan"))
     else:
         self.logger.warning("No kernel modules were autodetected.")
 
@@ -169,9 +188,9 @@ def _find_kernel_image(self) -> Path:
 
     for prefix in ["vmlinuz", "linux", "bzImage"]:
         if kernel_path := search_prefix(prefix):
-            self.logger.info("Detected kernel image: %s" % (colorize(kernel_path, "cyan")))
+            self.logger.info("Detected kernel image: %s" % (c_(kernel_path, "cyan")))
             return kernel_path
-    raise DependencyResolutionError("Failed to find kernel image")
+    raise AutodetectError("Failed to find kernel image")
 
 
 def _get_kver_from_header(self) -> str:
@@ -181,10 +200,12 @@ def _get_kver_from_header(self) -> str:
     The version string can be up to 127 bytes long and is null-terminated.
     https://www.kernel.org/doc/html/v6.7/arch/x86/boot.html#the-real-mode-kernel-header
     """
-    from struct import unpack
-
     kernel_path = _find_kernel_image(self)
-    kver_offset = unpack("<h", kernel_path.read_bytes()[0x020E:0x0210])[0] + 512
+    try:
+        kver_offset = unpack("<h", kernel_path.read_bytes()[0x020E:0x0210])[0] + 512
+    except StructError as e:
+        raise AutodetectError(f"Failed to read kernel version offset from: {kernel_path}") from e
+
     header = kernel_path.read_bytes()[kver_offset : kver_offset + 127]
     # Cut at the first null byte, decode to utf-8, and split at the first space
     kver = header[: header.index(0)].decode("utf-8").split()[0]
@@ -199,7 +220,11 @@ def _process_kernel_version(self, kver: str) -> None:
     if not kmod_dir.exists():
         if self["no_kmod"]:
             return self.logger.warning("[%s] Kernel module directory does not exist, but no_kmod is set." % kver)
-        raise DependencyResolutionError(f"Kernel module directory does not exist for kernel: {kver}")
+        self.logger.error(f"Available kernel versions: {', '.join([d.name for d in Path('/lib/modules').iterdir()])}")
+        self.logger.info(
+            "If kernel modules are not installed, and not required, set `no_kmod = true` to skip this check."
+        )
+        raise ValidationError(f"Kernel module directory does not exist for kernel: {kver}")
 
     self.data["kernel_version"] = kver
     self.data["_kmod_dir"] = kmod_dir
@@ -211,8 +236,8 @@ def _handle_arch_kernel(self) -> None:
     try:
         cmd = self._run(["pacman", "-Qqo", kernel_path])
         if not self["out_file"]:
-            self["out_file"] = f"/boot/initramfs-{cmd.stdout.decode().strip()}.img"
-            self.logger.info("Setting out_file to: %s" % colorize(self["out_file"], "green", bold=True))
+            self["out_file"] = f"{kernel_path.resolve().parent}/initramfs-{cmd.stdout.decode().strip()}.img"
+            self.logger.info("Setting out_file to: %s" % c_(self["out_file"], "green", bold=True))
     except RuntimeError as e:
         raise DependencyResolutionError("Failed to check ownership of kernel module directory") from e
 
@@ -237,7 +262,7 @@ def get_kernel_version(self) -> None:
             self["kernel_version"] = cmd.stdout.decode("utf-8").strip()
         except DependencyResolutionError:
             self["kernel_version"] = _get_kver_from_header(self)
-    self.logger.info("Detected kernel version: %s" % colorize(self["kernel_version"], "magenta", bold=True))
+    self.logger.info("Detected kernel version: %s" % c_(self["kernel_version"], "magenta", bold=True))
 
 
 @contains("kmod_init", "kmod_init is empty, skipping.")
@@ -263,15 +288,25 @@ def regen_kmod_metadata(self) -> None:
 
 
 def _add_kmod_firmware(self, kmod: str) -> None:
-    """Adds firmware files for the specified kernel module to the initramfs."""
+    """Adds firmware files for the specified kernel module to the initramfs.
+
+    Attempts to run even if no_kmod is set; this will not work if there are no kmods/no kernel version set
+    """
     kmod = _normalize_kmod_name(kmod)
+
     if kmod not in self["_kmod_modinfo"]:
+        if self["no_kmod"]:
+            return self.logger.warning(
+                "[%s] Kernel module info for firmware detection does not exist, but no_kmod is set." % kmod
+            )
         raise DependencyResolutionError("Kernel module info does not exist: %s" % kmod)
 
     if self["_kmod_modinfo"][kmod].get("firmware") and not self["kmod_pull_firmware"]:
+        # Log a warning if the kernel module has firmware files, but kmod_pull_firmware is not set
         self.logger.warning("[%s] Kernel module has firmware files, but kmod_pull_firmware is not set." % kmod)
 
     if not self["_kmod_modinfo"][kmod].get("firmware") or not self.get("kmod_pull_firmware"):
+        # No firmware files to add, or kmod_pull_firmware is not set
         return
 
     for firmware in self["_kmod_modinfo"][kmod]["firmware"]:
@@ -295,16 +330,21 @@ def _add_firmware_dep(self, kmod: str, firmware: str) -> None:
     self["dependencies"] = firmware_path
 
 
-def _process_kmod_dependencies(self, kmod: str) -> None:
-    """Processes a kernel module's dependencies."""
+def _process_kmod_dependencies(self, kmod: str, mod_tree=None) -> None:
+    """Processes a kernel module's dependencies.
+
+    If the kernel module is built in, only add firmware, don't resolve dependencies.
+
+    Only add softdeps if kmod_ignore_softdeps is not set.
+
+    Iterate over dependencies, adding them to kernel_mdules if they (or sub-dependencies) are not in the ignore list.
+    If the dependency is already in the module tree, skip it to prevent infinite recursion.
+    """
+    mod_tree = mod_tree or set()
     kmod = _normalize_kmod_name(kmod)
     _get_kmod_info(self, kmod)
 
-    if self["_kmod_modinfo"][kmod]["filename"] == "(builtin)":  # for built-in modules, just add firmware and return
-        _add_kmod_firmware(self, kmod)
-        raise BuiltinModuleError("Not adding built-in module to dependencies: %s" % kmod)
-
-    # Add dependencies of the module
+    # Get kernel module dependencies, softedeps if not ignored
     dependencies = []
     if harddeps := self["_kmod_modinfo"][kmod].get("depends"):
         dependencies += harddeps
@@ -315,7 +355,12 @@ def _process_kmod_dependencies(self, kmod: str) -> None:
         else:
             dependencies += sofdeps
 
+    # Iterate over module dependencies, skipping them if they are already in the mod_tree (to prevent infinite recursion)
+    # If the module isn't builtin or ignored, add it to the kernel_modules list and process its dependencies
     for dependency in dependencies:
+        if dependency in mod_tree:
+            self.logger.debug("[%s] Dependency is already in mod_tree: %s" % (kmod, dependency))
+            continue
         if dependency in self["kmod_ignore"]:  # Don't add modules with ignored dependencies
             _get_kmod_info(self, dependency)  # Make sure modinfo is queried in case it's built-in
             if modinfo := self["_kmod_modinfo"].get(dependency):
@@ -327,28 +372,49 @@ def _process_kmod_dependencies(self, kmod: str) -> None:
         if dependency in self["kernel_modules"]:
             self.logger.debug("[%s] Dependency is already in kernel_modules: %s" % (kmod, dependency))
             continue
+        mod_tree.add(dependency)
         try:
             self.logger.debug("[%s] Processing dependency: %s" % (kmod, dependency))
-            _process_kmod_dependencies(self, dependency)
+            _process_kmod_dependencies(self, dependency, mod_tree)
         except BuiltinModuleError as e:
             self.logger.debug(e)
             continue
         self["kernel_modules"] = dependency
 
-    # Process firmware now that dependencies are resolved
-    _add_kmod_firmware(self, kmod)
+    if self["_kmod_modinfo"][kmod]["filename"] == "(builtin)":  # for built-in modules, just add firmware and return
+        _add_kmod_firmware(self, kmod)
+        raise BuiltinModuleError("Not adding built-in module to dependencies: %s" % kmod)
 
-    # Add the kmod file to the initramfs dependenceis
-    filename = self["_kmod_modinfo"][kmod]["filename"]
-    if filename.endswith(".ko"):
-        self["dependencies"] = filename
-    elif filename.endswith(".ko.xz"):
-        self["xz_dependencies"] = filename
-    elif filename.endswith(".ko.gz"):
-        self["gz_dependencies"] = filename
-    else:
-        self.logger.warning("[%s] Unknown kmod extension: %s" % (kmod, filename))
-        self["dependencies"] = filename
+
+def add_kmod_deps(self):
+    """Adds all kernel modules to the initramfs dependencies.
+    Always attempt to add firmware, continuing if no_kmod is set.
+    If they are compressed with a supported extension, they are decompressed before being added.
+    """
+    for kmod in self["kernel_modules"]:
+        if self.get("kernel_version"):
+            _add_kmod_firmware(self, kmod)
+        else:
+            self.logger.warning(
+                f"Kernel version is not set, skipping firmware detection for kmod: {c_(kmod, 'yellow')}"
+            )
+        # if no_kmod is set, continue and check for the firmware of the next module
+        if self["no_kmod"]:
+            continue
+
+        # Add the kmod file to the initramfs dependenceis
+        filename = self["_kmod_modinfo"][kmod]["filename"]
+        if filename.endswith(".ko"):
+            self["dependencies"] = filename
+        elif filename.endswith(".ko.xz"):
+            self["xz_dependencies"] = filename
+        elif filename.endswith(".ko.zstd") or filename.endswith(".ko.zst"):
+            self["zstd_dependencies"] = filename
+        elif filename.endswith(".ko.gz"):
+            self["gz_dependencies"] = filename
+        else:
+            self.logger.warning("[%s] Unknown kmod extension: %s" % (kmod, filename))
+            self["dependencies"] = filename
 
 
 def process_ignored_module(self, module: str) -> None:
@@ -359,8 +425,14 @@ def process_ignored_module(self, module: str) -> None:
             if key == "kmod_init":
                 if module in self["_kmod_modinfo"] and self["_kmod_modinfo"][module]["filename"] == "(builtin)":
                     self.logger.debug("Removing built-in module from kmod_init: %s" % module)
+                elif module == "zfs":
+                    self.logger.critical("ZFS module is required but missing.")
+                    self.logger.critical("Please build/install the required kmods before running this script.")
+                    self.logger.critical("Detected kernel version: %s" % self["kernel_version"])
+                    # https://github.com/projg2/installkernel-gentoo/commit/1c70dda8cd2700e5306d2ed74886b66ad7ccfb42
+                    exit(77)
                 else:
-                    raise ValueError("Required module cannot be imported and is not builtin: %s" % module)
+                    raise MissingModuleError("Required module cannot be imported and is not builtin: %s" % module)
             else:
                 self.logger.debug("Removing ignored kernel module from %s: %s" % (key, module))
             self[key].remove(module)
@@ -373,9 +445,28 @@ def process_ignored_modules(self) -> None:
         process_ignored_module(self, module)
 
 
+def _process_optional_modules(self) -> None:
+    """Processes optional kernel modules."""
+    for kmod in self["kmod_init_optional"]:
+        if kmod in self["kmod_init"]:
+            self.logger.debug(f"Optional kmod_init module is already in kmod_init: {c_(kmod, 'yellow', bold=True)}")
+            continue
+        try:
+            _process_kmod_dependencies(self, kmod)
+            self["kmod_init"] = kmod  # add to kmod_init so it will be loaded
+        except IgnoredModuleError as e:
+            self.logger.warning(e)
+        except BuiltinModuleError:
+            self.logger.debug(f"Optional kmod_init module is built-in, skipping: {c_(kmod, 'yellow')}")
+            continue
+        except DependencyResolutionError as e:
+            self.logger.warning(f"[{c_(kmod, 'yellow', bold=True)}] Failed to process optional kernel module dependencies: {e}")
+
+
 @unset("no_kmod", "no_kmod is enabled, skipping.", log_level=30)
 def process_modules(self) -> None:
     """Processes all kernel modules, adding dependencies to the initramfs."""
+    _process_optional_modules(self)
     self.logger.debug("Processing kernel modules: %s" % self["kernel_modules"])
     for kmod in self["kernel_modules"].copy():
         self.logger.debug("Processing kernel module: %s" % kmod)
@@ -390,10 +481,11 @@ def process_modules(self) -> None:
             self["kernel_modules"].remove(kmod)
             continue  # Don't add built-in modules to the ignore list
         except IgnoredModuleError as e:
-            self.logger.info(e)
+            self.logger.warning(e)
         except DependencyResolutionError as e:
             if kmod in self["kmod_init"]:
-                self.logger.warning("[%s] Failed to get modinfo for init kernel module: %s" % (kmod, e))
+                # Once optional modules are fully implemented, this should raise an exception instead
+                self.logger.error("[%s] Failed to get modinfo for init kernel module: %s" % (kmod, e))
             self.logger.debug("[%s] Failed to get modinfo for kernel module: %s" % (kmod, e))
         self["kmod_ignore"] = kmod
 
@@ -409,8 +501,10 @@ def process_modules(self) -> None:
         except BuiltinModuleError:
             continue  # Don't add built-in modules to the ignore list
         except IgnoredModuleError as e:
-            self.logger.debug(e)
-        except DependencyResolutionError as e:
+            self.logger.debug(e)  # when autodetected modules are ignored, only debug log
+        except (
+            DependencyResolutionError
+        ) as e:  # log a warning, not error or exception when autodetected modules have missing deps
             self.logger.warning("[%s] Failed to process autodetected kernel module dependencies: %s" % (kmod, e))
         self["kmod_ignore"] = kmod
 
@@ -418,13 +512,24 @@ def process_modules(self) -> None:
 @contains("kmod_init", "No kernel modules to load.", log_level=30)
 def load_modules(self) -> str:
     """Creates a shell function which loads all kernel modules in kmod_init."""
-    self.logger.info(
-        "Init kernel modules: %s" % colorize(", ".join(self["kmod_init"]), "magenta", bright=True, bold=True)
-    )
-    if included_kmods := list(set(self["kernel_modules"]) ^ set(self["kmod_init"])):
-        self.logger.info("Included kernel modules: %s" % colorize(", ".join(included_kmods), "magenta"))
-    if removed_kmods := self.get("_kmod_removed"):
-        self.logger.warning("Ignored kernel modules: %s" % colorize(", ".join(removed_kmods), "red", bold=True))
+    init_kmods = ", ".join(self["kmod_init"])
+    included_kmods = ", ".join(list(set(self["kernel_modules"]) ^ set(self["kmod_init"])))
+    removed_kmods = ", ".join(self["_kmod_removed"])
+    if self["no_kmod"]:
+        if included_kmods or init_kmods:
+            self.logger.warning(
+                "no_kmod is enabled, but kernel modules are set, ensure the following kernel modules are built into the kernel:"
+            )
+            self.logger.warning(f"Init kernel modules: {c_(init_kmods, 'red', bold=True)}")
+            self.logger.warning(f"Included kernel modules: {c_(included_kmods, 'red', bold=True)}")
+        return
+
+    self.logger.info(f"Init kernel modules: {c_(init_kmods, 'magenta', bright=True, bold=True)}")
+    if included_kmods:
+        self.logger.info(f"Included kernel modules: {c_(included_kmods, 'magenta')}")
+
+    if removed_kmods:
+        self.logger.warning("Ignored kernel modules: %s" % c_(removed_kmods, "red", bold=True))
 
     module_list = " ".join(self["kmod_init"])
     return f"""
