@@ -1,6 +1,7 @@
 from unittest import TestCase, main
 
 from ugrd.initramfs_generator import InitramfsGenerator
+from ugrd.kmod import MissingModuleError
 from zenlib.logging import loggify
 
 
@@ -18,6 +19,13 @@ class TestCpio(TestCase):
         generator = InitramfsGenerator(logger=self.logger, config="tests/fs/xfs.toml")
         generator.build()
 
+    def test_f2fs(self):
+        generator = InitramfsGenerator(logger=self.logger, config="tests/fs/f2fs.toml")
+        try:
+            generator.build()
+        except MissingModuleError:
+            generator.logger.critical("F2FS is not supported on this system, skipping test.")
+
     def test_overlayfs(self):
         generator = InitramfsGenerator(logger=self.logger, config="tests/fs/overlayfs.toml")
         generator.build()
