@@ -1,4 +1,4 @@
-from unittest import TestCase, expectedFailure, main
+from unittest import TestCase, main
 
 from ugrd.initramfs_generator import InitramfsGenerator
 from zenlib.logging import loggify
@@ -14,9 +14,18 @@ class TestUGRD(TestCase):
         generator = InitramfsGenerator(logger=self.logger, config="tests/fullauto.toml", cpio_compression="xz")
         generator.build()
 
-    @expectedFailure
+    def test_zstd(self):
+        generator = InitramfsGenerator(logger=self.logger, config="tests/fullauto.toml", cpio_compression="zstd")
+        generator.build()
+
     def test_bad_config(self):
-        generator = InitramfsGenerator(logger=self.logger, config="tests/bad_config.toml")
+        with self.assertRaises(ValueError):
+            InitramfsGenerator(logger=self.logger, config="tests/bad_config.toml")
+
+    def test_no_root(self):
+        generator = InitramfsGenerator(
+            logger=self.logger, config="tests/fullauto.toml", test_no_rootfs=True, test_flag="Restarting init"
+        )
         generator.build()
 
 
